#
# logholder.rb
#   -- RICA event cacheing library
#   NISHI Takao <zophos@koka-in.org>
#
# $Id: logholder.rb,v 1.1 2001/08/29 05:11:16 zophos Exp $
#
require 'rica'

module Rica
    
    ############################################################
    #
    # Irc Log Holding Modules
    #
    module LogHolder

	########################################################
	#
	# Channel mode cache class
	#   Some non-RFC1459 mode supported.
	#   (I, B and E)
	#
	class ChannelMode
	    include Event

	    MODE_LINE=['o','p','s','i','t','n','m',
		'l','b','v','k','I','B','E']

	    ####################################################
	    #
	    # constructor
	    #
	    # Args:
	    #   Array of ChannelUser
	    #
	    def initialize(names)
		clear
		@names=names
	    end

	    ####################################################
	    #
	    # to_s override
	    #
	    # Return:
	    #   IRC message like strings. (eg. +stnl 10)
	    #
	    def to_s
		#
		# +o, +b, +v, +I, +B and +E modes are
		# ignored
		#
		modes=[nil,@p,@s,@i,@t,@n,@m,
		    @l,nil,@v,@k,nil,nil,nil]

		s=["+"]

		MODE_LINE.each{|c|
		    if(modes.shift)
			s[0]+=c
		    end
		}
		s.push(@l)
		s.push(@k)

		return s.join(" ").strip
	    end

	    alias put dispatch
	    
	    ####################################################
	    #
	    # MODE command
	    #
	    def on_recv_cmnd_mode(msg)
		purse(msg.args)
	    end

	    ####################################################
	    #
	    # RPL_CHANNELMODEIS (324)
	    #
	    def on_recv_channelmodeis(msg)
		clear
		purse(msg.args[2..-1])
	    end

	    private

	    ####################################################
	    #
	    # all mode clear
	    #
	    def clear
		@o=nil
		@p=false
		@s=false
		@i=false
		@t=false
		@n=false
		@m=false
		@l=nil
		@b=nil
		@v=nil
		@k=nil
		@I=nil
		@B=nil
		@E=nil
	    end

	    ####################################################
	    #
	    # purse mode strings
	    #
	    def purse(modes)
		flag=false

		while(modes)
		    mchars=modes.shift.split('')
		    mchars.each{|c|
			case c
			when '+'
			    flag=true
			when '-'
			    flag=false
			when 'o'
			    nick=modes.shift
			    i=@names.index(nick)
			    begin
				if(flag)
				    @names[i].setOp
				else
				    @names[i].deOp
				end
			    rescue
			    end
			when 'p'
			    @p=flag
			when 's'
			    @p=flag
			when 'i'
			    @i=flag
			when 't'
			    @t=flag
			when 'n'
			    @n=flag
			when 'm'
			    @m=flag
			when 'l'
			    if(flag)
				@l=modes.shift
			    else
				@l=nil
			    end
			when 'b'
			    modes.shift
			when 'v'
			    nick=modes.shift
			    i=@names.index(nick)
			    begin
				if(flag)
				    @names[i].setVoice
				else
				    @names[i].deVoice
				end
			    rescue
			    end
			when 'k'
			    key=modes.shift
			    if(flag)
				@k=key
			    else
				@k=nil
			    end
			when 'I'
			    modes.shift
			when 'B'
			    modes.shift
			when 'E'
			    modes.shift
			end
		    }
		end
	    end
	end

	########################################################
	#
	# User mode cache class
	#   A non-RFC1459 mode 'f' supported.
	#
	class UserMode
	    include Event
	    
	    MODE_LINE=['i','s','w','o','f']

	    def initialize
		clear
	    end

	    ####################################################
	    #
	    # to_s override
	    #
	    # Return:
	    #   IRC message like strings. (eg. +if)
	    #
	    def to_s
		s='+'
		modes=[@i,@s,@w,@o,@f]
		
		MODE_LINE.each{|m|
		    if(modes.shift)
			s+=m
		    end
		}

		return s
	    end

	    alias put dispatch

	    ####################################################
	    #
	    # MODE command
	    #
	    def on_recv_cmnd_mode(msg)
		purse(msg)
	    end

	    ####################################################
	    #
	    # RPL_UMODEIS (221)
	    #
	    def on_recv_rpl_umodeis(msg)
		clear
		purse(msg.args)
	    end

	    private

	    ####################################################
	    #
	    # all mode clear
	    #
	    def clear
		@i=false
		@s=false
		@w=false
		@o=false
		@f=false
	    end

	    ####################################################
	    #
	    # purse mode string
	    #
	    def purse(modes)
		flag=false

		modes=modes.join('').split('')
		modes.each{|m|
		    case m
		    when '+'
			flag=true
		    when '-'
			flag=false
		    when 'i'
			@i=flag
		    when 's'
			@s=flag
		    when 'w'
			@w=flag
		    when 'o'
			@o=flag
		    when 'f'
			@f=flag
		    end
		}
	    end
	end

	########################################################
	#
	# Channel user Informations
	#
	# Nick and two mode(+o,+v) is kept.
	#
	class ChannelUserInfo
	    include Event

	    ####################################################
	    #
	    # constructor
	    #
	    # Args:
	    #   String nick
	    #
	    def initialize(nick)
		@o=false
		@v=false
		
		case nick[0]
		when '@'
		    @o=true
		    nick=nick[1..-1]
		when '+'
		    @v=true
		    nick=nick[1..-1]
		end
		@nick=nick
	    end

	    attr_reader :nick
	    
	    ####################################################
	    #
	    # to_s override
	    #
	    # Return:
	    #   IRC message like strings. (eg. @hoge)
	    #
	    def to_s
		s=""
		if(@o)
		    s="@"
		elsif(@v)
		    s="+"
		end

		return s+@nick
	    end

	    ####################################################
	    #
	    # == override
	    #
	    # Compare a object and me using @nick without case
	    #
	    # This method use for Array#index or Array#delete. 
	    #
	    def ==(obj)
		begin
		    @nick.downcase==obj.downcase
		rescue NameError
		    @nick==obj
		end
	    end

	    ####################################################
	    #
	    # mode prefix
	    #
	    # Return :
	    #   String mode_prefix ('@'|'+'|'')
	    #
	    def mode
		if(@o)
		    "@"
		elsif(@v)
		    "+"
		else
		    ""
		end
	    end

	    def isOp?
		return(@o)
	    end

	    def isVoice?
		return(@o|@v)
	    end

	    def setOp
		@o=true
	    end

	    def setVoice
		@v=true
	    end

	    def deOp
		@o=false
		@v=false
	    end

	    def deVoice
		@v=false
	    end

	    alias put dispatch

	    def on_recv_cmnd_nick(msg)
		if(@nick==msg.fromNick)
		    @nick=msg.to
		end
	    end
	end
	
	########################################################
	#
	# Channel Informations
	#
	class ChannelInfo
	    include Event

	    ####################################################
	    #
	    # constructor
	    #
	    # Args:
	    #   server      :server name
	    #   channel     :channel name
	    #   conlose     :(true=server global information|
	    #                   false=channle local information)
	    #   logging_obj :logging object
	    #
	    # Note:
	    #   To logging, the logging_obj requires push() method.
	    #   If there is not push() method, log is not held.
	    #
	    def initialize(server,channel,console=false,logging_obj=nil)
		@server=server
		@channel=channel
		@topic=""
		@names=[]

		#
		# is server global information?
		#
		@console=console
		
		#
		# if self keep server global information,
		# @mode keeps UserMode, else @mode keeps
		# ChannelMode
		#
		if(@console)
		    @mode=UserMode.new
		else
		    @mode=ChannelMode.new(@names)
		end

		@log=logging_obj
	    end
	    
	    attr_reader :server
	    attr_reader :channel
	    attr_reader :topic
	    attr_reader :log

	    ####################################################
	    #
	    # Channel user list
	    #
	    # Args:
	    #   prefix :additional prefix
	    #   suffix :additional suffix
	    #
	    # Return:
	    #   String nameslist like as ircd names reply
	    #
	    # Note:
	    #   When prefix or suffix given, each nick is added
            #   prefix or suffix.
	    #
	    def names(prefix="",suffix="")
		if(prefix.empty? && suffix.empty?)
		    @names.join(" ")
		else
		    s=[]
		    @names.each{|n|
			s.push(n.mode+prefix+n.nick+suffix)
		    }
		    s.join(" ")
		end
	    end

	    ####################################################
	    #
	    # Channel or user mode
	    #
	    # Return:
	    #   String channel or user mode
	    #
	    def mode
		@mode.to_s
	    end

	    def put(msg)
		#
		# logging
		#
		begin
		    @log.push(msg)
		rescue NameError
		end

		dispatch(msg)
	    end

	    def on_recv_cmnd_nick(msg)
		i=@names.index(msg.fromNick)
		unless(i.nil?)
		    @names[i].put(msg)
		end
	    end

	    def on_recv_cmnd_quit(msg)
		@names.delete(msg.fromNick)
	    end

	    def on_recv_cmnd_join(msg)
		@names.unshift(ChannelUserInfo.new(msg.fromNick))
	    end
	    
	    def on_recv_cmnd_part(msg)
		@names.delete(msg.fromNick)
	    end

	    def on_recv_cmnd_mode(msg)
		@mode.put(msg)
	    end

	    def on_recv_cmnd_topic(msg)
		@topic=msg.args[0]
	    end

	    def on_recv_cmnd_kick(msg)
		@names.delete(msg.fromNick)
	    end

	    def on_recv_rpl_umodeis(msg)
		if(@console)
		    @mode.put(msg)
		end
	    end

	    def on_recv_rpl_channelmodeis(msg)
		unless(@console)
		    @mode.set(msg)
		end
	    end

	    def on_recv_rpl_notopic(msg)
		@topic=""
	    end

	    def on_recv_rpl_topic(msg)
		@topic=msg.args[1]
	    end

	    def on_recv_namereply(msg)
		@names.clear
		msg.args[2..-1].each{|nick|
		    @names.push(ChannelUserInfo.new(nick))
		}
	    end
	end

	########################################################
	#
	# Server Informations
	#
	class ServerInfo
	    include Event
	    
	    SERVER_COMMON_BUF=""


	    ####################################################
	    #
	    # constructor
	    #
	    # Args:
	    #   server        :server_name
	    #   logging_class :logging class
            #   logging_opt   :arguments for logging_class.new
	    #
	    def initialize(server=nil,logging_class=nil,*logging_opt)
		@server=server
		@channels={}
		@logging_class=logging_class
		@logging_opt=*logging_opt

		#
		# Create global information holder
		#
		begin
		    @channels[SERVER_COMMON_BUF]=
			ChannelInfo.new(@server,
					SERVER_COMMON_BUF,
					true,
					@logging_class.new(@logging_opt)
					)
		rescue NameError
		    @channels[SERVER_COMMON_BUF]=
			ChannelInfo.new(@server,
					SERVER_COMMON_BUF,
					true,
					nil
					)
		end
	    end
	    
	    attr_reader :channels

	    def [](channel)
		@channels[channel]
	    end

	    ####################################################
	    #
	    # get server global information holder
	    #
	    # Return :ChannelInfo server global messages
	    #
	    def console
		return @channels[SERVER_COMMON_BUF]
	    end

	    ####################################################
	    #
	    # get channel list
	    #
	    # Return :ChannelInfo server global messages
	    #
	    def channel_list
		tmp=@channels.keys
		tmp.delete(SERVER_COMMON_BUF)
		tmp
	    end

	    ####################################################
	    #
	    # purge channel information
	    #
	    def purge_channel(channel)
		@channels.delete(channel)
	    end

	    ####################################################
	    #
	    # get user mode
	    #
	    # Return :UserMode
	    #
	    def mode
		return @channels[SERVER_COMMON_BUF].mode
	    end

	    ####################################################
	    #
	    # Message dispatch methods
	    #
	    def put(msg)
		if(@server.nil?)
		    @server=msg.server
		end
		if(msg.server==@server)
		    dispatch(msg)
		end
	    end

	    ####################################################
	    #
	    # On defualt, given messges send to global
	    # information holder.
	    #
	    def default_action(msg)
		msg.add_info=SERVER_COMMON_BUF
		@channels[SERVER_COMMON_BUF].put(msg)
	    end
	    
	    ####################################################
	    #
	    # NICK and QUIT command messages are send to
	    # every holders.
	    #
	    alias on_recv_cmnd_nick broadcast
	    alias on_recv_cmnd_quit broadcast

	    def on_recv_cmnd_join(msg)
		put_to(msg.to,msg)
	    end
	    
	    def on_recv_cmnd_part(msg)
		put_to(msg.to,msg)
	    end
	    
	    def on_recv_cmnd_mode(msg)
		#
		# if usermode changed, the message is send to
		# global information holder.
		#
		if(msg.to==msg.selfNick)
		    @channels[SERVER_COMMON_BUF].put(msg)
		else
		    put_to(msg.to,msg)
		end
	    end

	    def on_recv_cmnd_topic(msg)
		put_to(msg.to,msg)
	    end

	    def on_recv_cmnd_kick(msg)
		put_to(msg.to,msg)
	    end

	    def on_recv_cmnd_privmsg(msg)
		#
		# When recept Priv, the message goes
		# to sender's holder.
		#
		if(msg.isPriv? && !msg.isSelfMessage?)
		    put_to(msg.fromNick,msg)
		else
		    put_to(msg.to,msg)
		end
	    end
	    alias on_recv_cmnd_notice on_recv_cmnd_privmsg
	    
	    def on_recv_rpl_channelmodeis(msg)
		put_to(msg.args[0],msg)
	    end

	    def on_recv_rpl_notopic(msg)
		put_to(msg.args[0],msg)
	    end

	    def on_recv_rpl_topic(msg)
		put_to(msg.args[0],msg)
	    end

	    #def on_recv_rpl_inviting(msg)
	    #end
	    
	    def on_recv_namereply(msg)
		put_to(msg.args[1],msg)
	    end
	    
	    def on_recv_endofname(msg)
		put_to(msg.args[0],msg)
	    end

	    def on_recv_rpl_banlist(msg)
		put_to(msg.args[0],msg)
	    end
	    
	    def on_recv_rpl_endofbanlist(msg)
		put_to(msg.args[0],msg)
	    end

	    def on_recv_err_cannotsendtochan(msg)
		put_to(msg.args[0],msg)
	    end

	    def on_recv_err_chanopprivsneeded(msg)
		put_to(msg.args[0],msg)
	    end

	    private
	    
	    ####################################################
	    #
	    # broadcast to every holders.
	    #
	    def broadcast(msg)
		@channels.each{|cn,obj|
		    #
		    # channel name is saved to Message.add_info
		    #
		    msg.add_info=cn
		    obj.put(msg)
		}
	    end

	    ####################################################
	    #
	    # send message to specified holder.
	    #
	    def put_to(channel,msg)

		#
		# If given channel information holder does
		# not exist, create it first.
		#
		unless(@channels.has_key?(channel))
		    begin
			@channels[channel]=
			    Channel.new(@server,
					channel,
					false,
					@logging_class.new(@logging_opt)
					)
		    rescue NameError
			@channels[channel]=
			    Channel.new(@server,
					channel,
					false,
					nil
					)
		    end
		end

		#
		# channel name is saved to Message.add_info
		#
		msg.add_info=channel
		@channels[channel].put(msg)
	    end

	end

	class Logger<MessageProcessor
	    def initialize(logging_class,*logging_opt)
		super
		@servers={}
		@logging_class=logging_class
		@logging_opt=*logging_opt
	    end
	    
	    attr_reader :servers

	    def [](server)
		@servers[server]
	    end

	    def default_action(msg)
		unless(@servers.has_key?(msg.server))
		    @servers[msg.server]=
			ServerInfo.new(
				       msg.server,
				       @logging_class,
				       @logging_opt
				       )
		end
		@servers[msg.server].put(msg)
	    end
	end
    end

end
